
# coding: utf-8

# # Decoding
# 
# If a notebook is imported, it should provide a natural __traceback__ experience similar to python imports.  The `importnb` importer creates a just decoder object that retains line numbers to the raw json when the notebook is imported.

# In[1]:


from json.decoder import JSONObject, JSONDecoder, WHITESPACE, WHITESPACE_STR    
from nbformat import NotebookNode
class LineNoDecoder(JSONDecoder):
    """A JSON Decoder to return a NotebookNode with lines numbers in the metadata."""
    def __init__(self, *, object_hook=None, parse_float=None, parse_int=None, parse_constant=None, strict=True, object_pairs_hook=None):
        from json.scanner import py_make_scanner    
        super().__init__(object_hook=object_hook, parse_float=parse_float, parse_int=parse_int, parse_constant=parse_constant, strict=strict, 
                         object_pairs_hook=object_pairs_hook)
        self.parse_object = self.object
        self.scan_once = py_make_scanner(self)
        
    def object(
        self, 
        s_and_end, 
        strict, scan_once, object_hook, object_pairs_hook, memo=None, _w=WHITESPACE.match, _ws=WHITESPACE_STR
    ) -> (NotebookNode, int):
        object, next = JSONObject(s_and_end, strict, scan_once, object_hook, object_pairs_hook, memo=memo, _w=_w, _ws=_ws)

        if 'cell_type' in object: 
            object['metadata'].update({
                'lineno':  len(s_and_end[0][:next].rsplit('"source":', 1)[0].splitlines())
            })
            
        for key in ('source', 'text'): 
            if key in object: object[key] = ''.join(object[key])
        return NotebookNode(object), next


# # Compilation
# 
# Compilation occurs in the __3__ steps:
# 
# 1. Text is transformed into a valid source string.
# 2. The source string is parsed into an abstract syntax tree
# 3. The abstract syntax compiles to valid bytecode 

# In[2]:


from IPython.core.compilerop import CachingCompiler
from dataclasses import dataclass, field

class Compiler(CachingCompiler):
    """{Shell} provides the IPython machinery to objects."""
    filename: str = '<Shell>'            
    @property
    def ip(Compiler): 
        """The current interactive shell"""
        from IPython import get_ipython
        from IPython.core.interactiveshell import InteractiveShell
        return get_ipython() or InteractiveShell()
    
    def ast_transform(Compiler, node):
        for visitor in Compiler.ip.ast_transformers: 
            node = visitor.visit(node)
        return node
    
    @property
    def transform(Compiler): return Compiler.ip.input_transformer_manager.transform_cell

    def compile(Compiler, ast): 
        """Compile AST to bytecode using the an IPython compiler."""
        return (Compiler.ip and Compiler.ip.compile or CachingCompiler())(ast, Compiler.filename, 'exec')
            
    def ast_parse(Compiler, source, filename='<unknown>', symbol='exec', lineno=0): 
        return ast.increment_lineno(super().ast_parse(source, Compiler.filename, 'exec'), lineno)


# In[3]:


import ast, sys
from json import load, loads
from nbformat import NotebookNode, read, reads
from pathlib import Path
from nbconvert.exporters.markdown import MarkdownExporter
from nbconvert.exporters.notebook import NotebookExporter


# In[4]:


@dataclass
class Code(NotebookExporter, Compiler):
    """An exporter than returns transforms a NotebookNode through the InputSplitter.
    
    >>> assert type(Code().from_filename('compiler.ipynb')) is NotebookNode"""
    filename: str = '<module exporter>'
    name: str = '__main__'
    decoder: type = LineNoDecoder
        
    def __post_init__(self): NotebookExporter.__init__(self) or Compiler.__init__(self)
            
    def from_file(Code,file_stream, resources=None, **dict): 
        for str in ('name', 'filename'): setattr(Code, str, dict.pop(str, getattr(Code, str)))
        return Code.from_notebook_node(
            NotebookNode(**load(file_stream, cls=Code.decoder)), resources, **dict)
    
    def from_filename(Code,  filename, resources=None, **dict):
        Code.filename, Code.name = filename, Path(filename).stem
        return super().from_filename(filename, resources, **dict)

    def from_notebook_node(Code, nb, resources=None, **dict): 
        for cell in nb['cells']:
            if cell['cell_type'] == 'code':
                cell.source = Code.from_code_cell(cell, **dict)
        return nb
    
    def from_code_cell(Code, cell, **dict):  
        return Code.transform(cell['source'])


# In[5]:


class AST(Code):
    """An exporter than returns parsed ast.
    
    >>> assert type(AST().from_filename('compiler.ipynb')) is ast.Module"""
    def from_notebook_node(AST, nb: NotebookNode, resource: dict=None, **dict):         
        return AST.ast_transform(ast.fix_missing_locations(ast.Module(body=sum((
            AST.ast_parse(
                AST.from_code_cell(cell, **dict), lineno=cell['metadata'].get('lineno', 1)
            ).body for cell in nb.cells if cell['cell_type']=='code'
        ), []))))


# In[6]:


class Compile(AST):
    """An exporter that returns compiled and cached bytecode.
    
    >>> assert Compile().from_filename('compiler.ipynb')"""        
    def from_notebook_node(Compile, nb, resources: dict=None, **dict):
        return Compile.compile(super().from_notebook_node(nb, resources, **dict))


# In[7]:


if __name__ ==  '__main__':
    get_ipython().system('jupyter nbconvert --to script compiler.ipynb')
    __import__('doctest').testmod()





# coding: utf-8

# # The [Import Loader](https://docs.python.org/3/reference/import.html#loaders)

# In[1]:


try:
    from .compiler import Compile, AST
except ModuleNotFoundError:
    from compiler import Compile, AST
import inspect, sys
from importlib.machinery import SourceFileLoader
from importlib._bootstrap_external import FileFinder
from importlib import reload
from traceback import print_exc


# In[2]:


def update_path_hooks(loader: SourceFileLoader, extensions: tuple=None, lazy=False):
    """Update the FileFinder loader in sys.path_hooks to accomodate a {loader} with the {extensions}"""
    from importlib.util import LazyLoader
    for id, hook in enumerate(sys.path_hooks):
        try:
            closure = inspect.getclosurevars(hook).nonlocals
        except TypeError: continue
        if issubclass(closure['cls'], FileFinder):
            sys.path_hooks.pop(id)
            sys.path_hooks.insert(id, FileFinder.path_hook(*(
                ((lazy and LazyLoader.factory(loader) or loader, extensions),) 
                if (loader and extensions) 
                else tuple()
            ) + tuple(
                (cls, ext) 
                for cls, ext in closure['loader_details'] 
                if not issubclass(cls, loader) # Need to add logic for lazy loaders before they may be introduced.
            )))
    sys.path_importer_cache.clear()


# In[3]:


class ImportContextMixin:
    def __enter__(self):  update_path_hooks(type(self), self.EXTENSION_SUFFIXES)
    def __exit__(self, exception_type=None, exception_value=None, traceback=None): update_path_hooks(type(self))


# In[4]:


class Notebook(SourceFileLoader, ImportContextMixin):
    """A SourceFileLoader for notebooks that provides line number debugginer in the JSON source."""
    EXTENSION_SUFFIXES = '.ipynb',
    
    def __init__(self, fullname=None, path=None): super().__init__(fullname, path)
    
    def exec_module(Loader, module):
        from IPython.utils.capture import capture_output    
        with capture_output(stdout=False, stderr=False) as output: 
            try: super().exec_module(module)
            except: ...
            finally: module.__output__ = output
        return module

    def source_to_code(Notebook, data, path):
        with __import__('io').BytesIO(data) as stream:
            return Compile().from_file(stream, filename=Notebook.path, name=Notebook.name)


# In[5]:


class Partial(Notebook):
    def exec_module(loader, module):
        try: super().exec_module(module)
        except BaseException as exception:
            try: raise ImportWarning("""{name} from {file} failed to load completely.""".format(
                name=module.__name__, file=module.__file__
            ))
            except ImportWarning as error:
                print_exc()
                module.__exception__ = exception
        return module


# # IPython Extensions

# In[6]:


def load_ipython_extension(ip=None): Notebook().__enter__()
def unload_ipython_extension(ip=None): Notebook().__exit__()


# In[7]:


class Test(__import__('unittest').TestCase): 
    def setUp(Test):
        from nbformat import write, v4
        load_ipython_extension()
        with open('test_loader.ipynb', 'w') as file:
            write(v4.new_notebook(cells=[
                v4.new_code_cell("""__import__("time").sleep(1);test = 42""")
            ]), file)
            
    def runTest(Test):
        import test_loader
        assert test_loader.__file__.endswith('.ipynb')
        assert test_loader.test is 42
        assert isinstance(test_loader, __import__('types').ModuleType)
        
    def tearDown(Test):
#             %rm test_loader.ipynb
        unload_ipython_extension()


# # Developer

# In[8]:


if __name__ ==  '__main__':
#         __import__('doctest').testmod(verbose=2)
#         __import__('unittest').TextTestRunner().run(Test())
    get_ipython().system('jupyter nbconvert --to script --TemplateExporter.exclude_input_prompt=True __init__.ipynb')
    with Notebook():
        import test_failure




from pathlib import Path
import setuptools

name = "importnb"

__version__ = None

here = Path(__file__).parent

exec((here / name / "_version.py").read_text())

setup_args = dict(
    name=name,
    version=__version__,
    author="deathbeds", 
    author_email="tony.fast@gmail.com",
    description="Import .ipynb files as modules in the system path.", 
    long_description=(
        (here / "readme.md").read_text() + "\n\n" +
        (here / "changelog.md").read_text()
    ),
    url="https://github.com/deathbeds/importnb",
    python_requires=">=3.6",
    license="BSD-3-Clause",
    install_requires=[
        "dataclasses",
        "nbconvert",
    ],
    include_package_data=True,
    packages=setuptools.find_packages(),
    classifiers=(
        "Development Status :: 4 - Beta",
        "Framework :: IPython",
        "Framework :: Jupyter",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
    ),
    zip_safe=False,
)

if __name__ == "__main__":
    setuptools.setup(**setup_args)